from yamcs.core import pagination
from yamcs.core.helpers import adapt_name_for_rest
from yamcs.mdb.model import Algorithm, Command, Container, Parameter, SpaceSystem
from yamcs.protobuf.mdb import mdb_pb2


class MDBClient:
    def __init__(self, ctx, instance):
        super(MDBClient, self).__init__()
        self.ctx = ctx
        self._instance = instance

    def list_space_systems(self, page_size=None):
        """
        Lists the space systems visible to this client.

        Space systems are returned in lexicographical order.

        :rtype: :class:`.SpaceSystem` iterator
        """
        params = {}

        if page_size is not None:
            params["limit"] = page_size

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/mdb/{self._instance}/space-systems",
            params=params,
            response_class=mdb_pb2.ListSpaceSystemsResponse,
            items_key="spaceSystems",
            item_mapper=SpaceSystem,
        )

    def get_space_system(self, name):
        """
        Gets a single space system by its unique name.

        :param str name: A fully-qualified XTCE name
        :rtype: .SpaceSystem
        """
        url = f"/mdb/{self._instance}/space-systems{name}"
        response = self.ctx.get_proto(url)
        message = mdb_pb2.SpaceSystemInfo()
        message.ParseFromString(response.content)
        return SpaceSystem(message)

    def list_parameters(self, parameter_type=None, page_size=None):
        """Lists the parameters visible to this client.

        Parameters are returned in lexicographical order.

        :param str parameter_type: The type of parameter
        :rtype: :class:`.Parameter` iterator
        """
        params = {"details": True}

        if parameter_type is not None:
            params["type"] = parameter_type
        if page_size is not None:
            params["limit"] = page_size

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/mdb/{self._instance}/parameters",
            params=params,
            response_class=mdb_pb2.ListParametersResponse,
            items_key="parameters",
            item_mapper=Parameter,
        )

    def get_parameter(self, name):
        """
        Gets a single parameter by its name.

        :param str name: Either a fully-qualified XTCE name or an alias in the
                         format ``NAMESPACE/NAME``.
        :rtype: .Parameter
        """
        name = adapt_name_for_rest(name)
        url = f"/mdb/{self._instance}/parameters{name}"
        response = self.ctx.get_proto(url)
        message = mdb_pb2.ParameterInfo()
        message.ParseFromString(response.content)
        return Parameter(message)

    def list_containers(self, page_size=None):
        """
        Lists the containers visible to this client.

        Containers are returned in lexicographical order.

        :rtype: :class:`.Container` iterator
        """
        params = {}

        if page_size is not None:
            params["limit"] = page_size

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/mdb/{self._instance}/containers",
            params=params,
            response_class=mdb_pb2.ListContainersResponse,
            items_key="containers",
            item_mapper=Container,
        )

    def get_container(self, name):
        """
        Gets a single container by its unique name.

        :param str name: Either a fully-qualified XTCE name or an alias in the
                         format ``NAMESPACE/NAME``.
        :rtype: .Container
        """
        name = adapt_name_for_rest(name)
        url = f"/mdb/{self._instance}/containers{name}"
        response = self.ctx.get_proto(url)
        message = mdb_pb2.ContainerInfo()
        message.ParseFromString(response.content)
        return Container(message)

    def list_commands(self, page_size=None):
        """
        Lists the commands visible to this client.

        Commands are returned in lexicographical order.

        :rtype: :class:`.Command` iterator
        """
        params = {}

        if page_size is not None:
            params["limit"] = page_size

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/mdb/{self._instance}/commands",
            params=params,
            response_class=mdb_pb2.ListCommandsResponse,
            items_key="commands",
            item_mapper=Command,
        )

    def get_command(self, name):
        """
        Gets a single command by its unique name.

        :param str name: Either a fully-qualified XTCE name or an alias in the
                         format ``NAMESPACE/NAME``.
        :rtype: .Command
        """
        name = adapt_name_for_rest(name)
        url = f"/mdb/{self._instance}/commands{name}"
        response = self.ctx.get_proto(url)
        message = mdb_pb2.CommandInfo()
        message.ParseFromString(response.content)
        return Command(message)

    def list_algorithms(self, page_size=None):
        """
        Lists the algorithms visible to this client.

        Algorithms are returned in lexicographical order.

        :rtype: :class:`.Algorithm` iterator
        """
        params = {}

        if page_size is not None:
            params["limit"] = page_size

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/mdb/{self._instance}/algorithms",
            params=params,
            response_class=mdb_pb2.ListAlgorithmsResponse,
            items_key="algorithms",
            item_mapper=Algorithm,
        )

    def get_algorithm(self, name):
        """
        Gets a single algorithm by its unique name.

        :param str name: Either a fully-qualified XTCE name or an alias in the
                         format ``NAMESPACE/NAME``.
        :rtype: .Algorithm
        """
        name = adapt_name_for_rest(name)
        url = f"/mdb/{self._instance}/algorithms{name}"
        response = self.ctx.get_proto(url)
        message = mdb_pb2.AlgorithmInfo()
        message.ParseFromString(response.content)
        return Algorithm(message)



from yamcs.core.helpers import parse_server_time
from yamcs.protobuf.events import events_pb2
from yamcs.protobuf.links import links_pb2
from yamcs.protobuf.yamcsManagement import yamcsManagement_pb2


class AuthInfo:
    """
    Authentication information
    """

    def __init__(self, proto):
        self._proto = proto

    @property
    def require_authentication(self):
        return self._proto.requireAuthentication


class ServerInfo:
    """
    General server properties.
    """

    def __init__(self, proto):
        self._proto = proto

    @property
    def id(self):
        """The Server ID."""
        return self._proto.serverId

    @property
    def version(self):
        """The version of Yamcs Server."""
        return self._proto.yamcsVersion

    @property
    def default_yamcs_instance(self):
        """Returns the default Yamcs instance."""
        if self._proto.HasField("defaultYamcsInstance"):
            return self._proto.defaultYamcsInstance
        return None

    def __str__(self):
        return f"{self.id} (v{self.version})"


class UserInfo:
    """
    Info on a Yamcs User.
    """

    def __init__(self, proto):
        self._proto = proto

    @property
    def username(self):
        return self._proto.name

    @property
    def superuser(self):
        return self._proto.superuser

    @property
    def system_privileges(self):
        return [p for p in self._proto.systemPrivilege]

    @property
    def object_privileges(self):
        return [ObjectPrivilege(p) for p in self._proto.objectPrivilege]

    def __str__(self):
        return self.username


class ObjectPrivilege:
    def __init__(self, proto):
        self._proto = proto

    @property
    def name(self):
        return self._proto.type

    @property
    def objects(self):
        return [o for o in self._proto.object]

    def __str__(self):
        return self.name


class Event:
    """
    A timetagged free-text message. Events work a lot like log
    messages in logging frameworks, but then targeted at operators.
    """

    def __init__(self, proto):
        self._proto = proto

    @property
    def generation_time(self):
        """
        The time when the event was generated.

        :type: :class:`~datetime.datetime`
        """
        if self._proto.HasField("generationTime"):
            return parse_server_time(self._proto.generationTime)
        return None

    @property
    def reception_time(self):
        """
        The time when the event was received by Yamcs.

        :type: :class:`~datetime.datetime`
        """
        if self._proto.HasField("receptionTime"):
            return parse_server_time(self._proto.receptionTime)
        return None

    @property
    def severity(self):
        """
        Severity level of the event. One of ``INFO``, ``WATCH``,
        ``WARNING``, ``DISTRESS``, ``CRITICAL`` or ``SEVERE``.
        """
        if self._proto.HasField("severity"):
            return events_pb2.Event.EventSeverity.Name(self._proto.severity)
        return None

    @property
    def message(self):
        """
        Event message.
        """
        if self._proto.HasField("message"):
            return self._proto.message
        return None

    @property
    def sequence_number(self):
        """
        Sequence number. Usually this is assigned by the source of the event.
        """
        if self._proto.HasField("seqNumber"):
            return self._proto.seqNumber
        return None

    @property
    def event_type(self):
        """
        The event type. This is mission-specific and can be any string.
        """
        if self._proto.HasField("type"):
            return self._proto.type
        return None

    @property
    def source(self):
        """
        The event source. Can be any string.
        """
        if self._proto.HasField("source"):
            return self._proto.source
        return None

    def __str__(self):
        return f"{self.generation_time} [{self.severity}] {self.message}"


class LinkEvent:
    """
    Data holder used in link subscriptions.
    """

    def __init__(self, proto):
        self._proto = proto

    @property
    def event_type(self):
        """
        The type of the event. One of ``REGISTERED``, ``UNREGISTERED``,
        or ``UPDATED``.
        """
        return links_pb2.LinkEvent.Type.Name(self._proto.type)

    @property
    def link(self):
        """
        Link state at the time of this event.

        :type: :class:`.Link`
        """
        return Link(self._proto.linkInfo)

    def __str__(self):
        return f"[{self.event_type}] {self.link}"


class Link:
    """
    Represents a link with an external system. Depending on the
    semantics of the link, this may imply inbound data, outbound
    data or a combination of both.
    """

    def __init__(self, proto):
        self._proto = proto

    @property
    def instance(self):
        """Name of the instance where this link is defined."""
        return self._proto.instance

    @property
    def name(self):
        """Name of this link (unique per instance)."""
        return self._proto.name

    @property
    def class_name(self):
        """Name of this link's class."""
        return self._proto.type

    @property
    def enabled(self):
        """If ``True``, this link accepts or outputs data."""
        return not self._proto.disabled

    @property
    def status(self):
        """Short status."""
        return self._proto.status

    @property
    def in_count(self):
        """The number of inbound data events (example: packet count)."""
        return self._proto.dataInCount

    @property
    def out_count(self):
        """The number of outbound data events (example: command count)."""
        return self._proto.dataOutCount

    def __str__(self):
        desc = f"{self.instance}/{self.name}"
        return f"{desc}: {self.status} (in: {self.in_count} out: {self.out_count})"


class Instance:
    def __init__(self, proto):
        self._proto = proto

    @property
    def name(self):
        """Name of this instance."""
        return self._proto.name

    @property
    def state(self):
        """
        State of this instance. One of ``OFFLINE``, ``INITIALIZING``,
        ``INITIALIZED``, ``STARTING``, ``RUNNING``, ``STOPPING`` or
        ``FAILED``.
        """
        if self._proto.HasField("state"):
            return yamcsManagement_pb2.YamcsInstance.InstanceState.Name(
                self._proto.state
            )
        return None

    @property
    def failure_cause(self):
        """Failure message when ``state == 'FAILED'``"""
        if self._proto.HasField("failureCause"):
            return self._proto.failureCause
        return None

    @property
    def mission_time(self):
        """
        Mission time of this instance's time service.

        :type: :class:`~datetime.datetime`
        """
        if self._proto.HasField("missionTime"):
            return parse_server_time(self._proto.missionTime)
        return None

    def __str__(self):
        return f"{self.name} [{self.state}]"


class InstanceTemplate:
    """A template for creating an instance."""

    def __init__(self, proto):
        self._proto = proto

    @property
    def name(self):
        """Name of this template."""
        return self._proto.name

    def __str__(self):
        return self.name


class Service:
    """A Yamcs service."""

    def __init__(self, proto):
        self._proto = proto

    @property
    def name(self):
        """Name of this service."""
        return self._proto.name

    @property
    def instance(self):
        """Name of the instance where this service is defined."""
        if self._proto.HasField("instance"):
            return self._proto.instance
        return None

    @property
    def processor(self):
        """Name of the processor where this service is defined."""
        if self._proto.HasField("processor"):
            return self._proto.processor
        return None

    @property
    def class_name(self):
        """Name of this service's class."""
        return self._proto.className

    @property
    def state(self):
        """State of this service."""
        if self._proto.HasField("state"):
            return yamcsManagement_pb2.ServiceState.Name(self._proto.state)
        return None

    def __str__(self):
        return f"{self.name} [{self.state}]"


class Processor:
    def __init__(self, proto):
        self._proto = proto

    @property
    def name(self):
        """Name of this processor."""
        return self._proto.name

    @property
    def instance(self):
        """Name of the instance where this processor is defined."""
        return self._proto.instance

    @property
    def state(self):
        """State of this processor."""
        if self._proto.HasField("state"):
            return yamcsManagement_pb2.ServiceState.Name(self._proto.state)
        return None

    @property
    def type(self):
        """Type of this processor."""
        return self._proto.type

    @property
    def owner(self):
        """User that owns this processor."""
        return self._proto.creator

    @property
    def persistent(self):
        """If ``True``, this processor does not close if no clients are connected."""
        return not self._proto.persistent

    @property
    def mission_time(self):
        """
        Mission time of this processor.

        :type: :class:`~datetime.datetime`
        """
        if self._proto.HasField("time"):
            return parse_server_time(self._proto.time)
        return None

    def __str__(self):
        return f"{self.name} [{self.state}]"



import abc
from xmlrpc.client import boolean

from yamcs.core.helpers import ProtoList, parse_server_time, to_server_time
from yamcs.protobuf.timeline import timeline_pb2


class View:
    def __init__(self, proto=None):
        merged = timeline_pb2.TimelineView()
        if proto:
            merged.MergeFrom(proto)
        self._proto = merged

        self._bands = ProtoList(self._proto, "bands", lambda x: Band._as_subclass(x))

    @property
    def id(self):
        """View identifier."""
        if self._proto.HasField("id"):
            return self._proto.id
        return None

    @property
    def name(self):
        """Name of this view."""
        return self._proto.name

    @name.setter
    def name(self, value):
        self._proto.name = value

    @property
    def description(self):
        """Description of this view."""
        return self._proto.description

    @description.setter
    def description(self, value):
        self._proto.name = value

    @property
    def bands(self):
        """
        Bands included in this view.

        :type: List[:class:`.Band`]
        """
        return self._bands

    @bands.setter
    def bands(self, value):
        self._bands.clear()
        self._bands.extend(value)

    def __str__(self):
        return self.name


class Item:
    def __init__(self, proto=None):
        merged = timeline_pb2.TimelineItem()
        merged.type = timeline_pb2.TimelineItemType.EVENT
        if proto:
            merged.MergeFrom(proto)
        self._proto = merged

    @property
    def id(self):
        """Item identifier."""
        if self._proto.HasField("id"):
            return self._proto.id
        return None

    @property
    def item_type(self):
        """Type of item."""
        return timeline_pb2.TimelineItemType.Name(self._proto.type)

    @property
    def name(self):
        """Name of this item."""
        return self._proto.name

    @name.setter
    def name(self, value):
        self._proto.name = value

    @property
    def tags(self):
        """
        Item tags. Used by bands to filter what is visible.

        :type: str[]
        """
        return self._proto.tags

    @tags.setter
    def tags(self, value):
        self._proto.tags[:] = value

    @property
    def start(self):
        """
        Item start time.

        :type: :class:`~datetime.datetime`
        """
        if self._proto.HasField("start"):
            return parse_server_time(self._proto.start)
        return None

    @start.setter
    def start(self, value):
        self._proto.start.MergeFrom(to_server_time(value))

    @property
    def duration(self):
        """
        Item duration.

        :type: :class:`~datetime.timedelta`
        """
        if self._proto.HasField("duration"):
            return self._proto.duration.ToTimedelta()
        return None

    @duration.setter
    def duration(self, value):
        self._proto.duration.FromTimedelta(value)

    def __str__(self):
        return self.name


class Band(abc.ABC):
    """
    Superclass for bands. Implementations:

    * :class:`.TimeRuler`
    * :class:`.ItemBand`
    * :class:`.Spacer`
    * :class:`.CommandBand`
    """

    def __init__(self, proto):
        self._proto = proto

    @property
    def id(self):
        """Band identifier."""
        if self._proto.HasField("id"):
            return self._proto.id
        return None

    @property
    def band_type(self):
        """Type of band."""
        return timeline_pb2.TimelineBandType.Name(self._proto.type)

    @property
    def name(self):
        """Name of this band."""
        return self._proto.name

    @name.setter
    def name(self, value):
        self._proto.name = value

    @property
    def description(self):
        """Description of this band."""
        return self._proto.description

    @description.setter
    def description(self, value):
        self._proto.description = value

    def _set_integer_property(self, key, value):
        if not isinstance(value, int):
            raise ValueError("Provided value is not integer")
        self._proto.properties[key] = str(value)

    def _get_integer_property(self, key):
        return int(self._proto.properties[key])

    def _set_boolean_property(self, key, value):
        if not isinstance(value, boolean):
            raise ValueError("Provided value is not boolean")
        self._proto.properties[key] = "true" if value else "false"

    def _get_boolean_property(self, key):
        return self._proto.properties[key] == "true"

    @staticmethod
    def _as_subclass(proto):
        if proto.type == timeline_pb2.TimelineBandType.TIME_RULER:
            return TimeRuler(proto)
        elif proto.type == timeline_pb2.TimelineBandType.ITEM_BAND:
            return ItemBand(proto)
        elif proto.type == timeline_pb2.TimelineBandType.SPACER:
            return Spacer(proto)
        elif proto.type == timeline_pb2.TimelineBandType.COMMAND_BAND:
            return CommandBand(proto)
        else:
            raise ValueError("Unexpected band type")

    def __str__(self):
        return self.name


class TimeRuler(Band):
    """
    Displays absolute time, formatted in a timezone of choice.
    """

    def __init__(self, proto=None):
        merged = timeline_pb2.TimelineBand()
        merged.type = timeline_pb2.TimelineBandType.TIME_RULER
        merged.properties["timezone"] = "UTC"
        if proto:
            merged.MergeFrom(proto)
        super(TimeRuler, self).__init__(merged)

    @property
    def timezone(self):
        """
        IANA timezone name.

        Corresponds with the third column of the following table:
        https://data.iana.org/time-zones/data/zone1970.tab

        In addition, the name `UTC` is supported.
        """
        return self._proto.properties["timezone"]

    @timezone.setter
    def timezone(self, value):
        self._proto.properties["timezone"] = value


class Spacer(Band):
    """
    Insert empty vertical space.
    """

    def __init__(self, proto=None):
        merged = timeline_pb2.TimelineBand()
        merged.type = timeline_pb2.TimelineBandType.SPACER
        merged.properties["height"] = "34"
        if proto:
            merged.MergeFrom(proto)
        super(Spacer, self).__init__(merged)

    @property
    def height(self):
        """Spacer height"""
        return self._get_integer_property("height")

    @height.setter
    def height(self, value):
        self._set_integer_property("height", value)


class CommandBand(Band):
    """
    Display issued commands.
    """

    def __init__(self, proto=None):
        merged = timeline_pb2.TimelineBand()
        merged.type = timeline_pb2.TimelineBandType.COMMAND_BAND
        if proto:
            merged.MergeFrom(proto)
        super(CommandBand, self).__init__(merged)


class ItemBand(Band):
    """
    Show a selection of timeline items.
    """

    def __init__(self, proto=None):
        merged = timeline_pb2.TimelineBand()
        merged.type = timeline_pb2.TimelineBandType.ITEM_BAND
        merged.properties["frozen"] = "false"
        merged.properties["itemBackgroundColor"] = "#77b1e1"
        merged.properties["itemBorderColor"] = "#3d94c7"
        merged.properties["itemBorderWidth"] = "1"
        merged.properties["itemCornerRadius"] = "0"
        merged.properties["itemHeight"] = "20"
        merged.properties["itemMarginLeft"] = "5"
        merged.properties["itemTextColor"] = "#333333"
        merged.properties["itemTextOverflow"] = "show"
        merged.properties["itemTextSize"] = "10"
        merged.properties["marginTop"] = "7"
        merged.properties["marginBottom"] = "7"
        merged.properties["multiline"] = "true"
        merged.properties["spaceBetweenItems"] = "0"
        merged.properties["spaceBetweenLines"] = "2"
        if proto:
            merged.MergeFrom(proto)
        super(ItemBand, self).__init__(merged)

    @property
    def frozen(self):
        """
        Fix this line to the top of the view. Frozen bands are always
        rendered above other bands.
        """
        return self._get_boolean_property("frozen")

    @frozen.setter
    def frozen(self, value):
        self._set_boolean_property("frozen", value)

    @property
    def tags(self):
        """
        Item tags that this band filters on.

        :type: str[]
        """
        return self._proto.tags

    @tags.setter
    def tags(self, value):
        self._proto.tags[:] = value

    @property
    def item_background_color(self):
        """CSS color string."""
        return self._proto.properties["itemBackgroundColor"]

    @item_background_color.setter
    def item_background_color(self, value):
        self._proto.properties["itemBackgroundColor"] = value

    @property
    def item_border_color(self):
        """CSS color string."""
        return self._proto.properties["itemBorderColor"]

    @item_border_color.setter
    def item_border_color(self, value):
        self._proto.properties["itemBorderColor"] = value

    @property
    def item_border_width(self):
        return self._get_integer_property("itemBorderWidth")

    @item_border_width.setter
    def item_border_width(self, value):
        self._set_integer_property("itemBorderWidth", value)

    @property
    def item_corner_radius(self):
        return self._get_integer_property("itemCornerRadius")

    @item_corner_radius.setter
    def item_corner_radius(self, value):
        self._set_integer_property("itemCornerRadius", value)

    @property
    def item_height(self):
        return self._get_integer_property("itemHeight")

    @item_height.setter
    def item_height(self, value):
        self._set_integer_property("itemHeight", value)

    @property
    def item_margin_left(self):
        return self._get_integer_property("itemMarginLeft")

    @item_margin_left.setter
    def item_margin_left(self, value):
        self._set_integer_property("itemMarginLeft", value)

    @property
    def item_text_color(self):
        """CSS color string."""
        return self._proto.properties["itemTextColor"]

    @item_text_color.setter
    def item_text_color(self, value):
        self._proto.properties["itemTextColor"] = value

    @property
    def item_text_overflow(self):
        """One of ``show``, ``clip``, or ``hide``."""
        return self._proto.properties["itemTextOverflow"]

    @item_text_overflow.setter
    def item_text_overflow(self, value):
        self._proto.properties["itemTextOverflow"] = value

    @property
    def item_text_size(self):
        return self._get_integer_property("itemTextSize")

    @item_text_size.setter
    def item_text_size(self, value):
        self._set_integer_property("itemTextSize", value)

    @property
    def margin_bottom(self):
        return self._get_integer_property("marginBottom")

    @margin_bottom.setter
    def margin_bottom(self, value):
        self._set_integer_property("marginBottom", value)

    @property
    def margin_top(self):
        return self._get_integer_property("marginTop")

    @margin_top.setter
    def margin_top(self, value):
        self._set_integer_property("marginTop", value)

    @property
    def multiline(self):
        """
        Draw items on multiple lines if otherwise there would be collisions.
        """
        return self._get_boolean_property("multiline")

    @multiline.setter
    def multiline(self, value):
        self._set_boolean_property("multiline", value)

    @property
    def space_between_items(self):
        """
        In case of multilining, this indicates the minimum horizontal space
        between items. If an item does not meet this treshold, it gets
        rendered on a different line.
        """
        return self._get_integer_property("spaceBetweenItems")

    @space_between_items.setter
    def space_between_items(self, value):
        self._set_integer_property("spaceBetweenItems", value)

    @property
    def space_between_lines(self):
        """
        In case of multilining, this indicates the vertical space between
        lines.
        """
        return self._get_integer_property("spaceBetweenLines")

    @space_between_lines.setter
    def space_between_lines(self, value):
        self._set_integer_property("spaceBetweenLines", value)



import functools
from datetime import datetime, timedelta, timezone

from yamcs.archive.model import (
    IndexGroup,
    ParameterRange,
    ResultSet,
    Sample,
    Stream,
    StreamData,
    Table,
)
from yamcs.core import pagination
from yamcs.core.futures import WebSocketSubscriptionFuture
from yamcs.core.helpers import (
    split_protobuf_stream,
    to_isostring,
    to_named_object_ids,
    to_server_time,
)
from yamcs.core.subscriptions import WebSocketSubscriptionManager
from yamcs.model import Event
from yamcs.protobuf.archive import (
    archive_pb2,
    index_service_pb2,
    parameter_archive_service_pb2,
)
from yamcs.protobuf.commanding import commands_service_pb2
from yamcs.protobuf.events import events_service_pb2
from yamcs.protobuf.packets import packets_pb2, packets_service_pb2
from yamcs.protobuf.pvalue import pvalue_pb2
from yamcs.protobuf.table import table_pb2
from yamcs.tmtc.model import CommandHistory, Packet, ParameterData, ParameterValue


def _wrap_callback_parse_stream_data(subscription, on_data, message):
    """
    Wraps an (optional) user callback to parse StreamData
    from a WebSocket data message
    """
    pb = table_pb2.StreamData()
    message.Unpack(pb)
    on_data(StreamData(pb))


class ArchiveClient:
    def __init__(self, ctx, instance):
        super(ArchiveClient, self).__init__()
        self.ctx = ctx
        self._instance = instance

    def list_packet_names(self):
        """
        Returns the existing packet names.

        :rtype: ~collections.abc.Iterable[str]
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/packet-names"
        response = self.ctx.get_proto(path=path)
        message = packets_service_pb2.ListPacketNamesResponse()
        message.ParseFromString(response.content)
        names = getattr(message, "name")
        return iter(names)

    def list_packet_histogram(self, name=None, start=None, stop=None, merge_time=2):
        """
        Reads packet-related index records between the specified start and stop
        time.

        Each iteration returns a chunk of chronologically-sorted records.

        :param float merge_time: Maximum gap in seconds before two consecutive index
                                 records are merged together.
        :rtype: ~collections.abc.Iterable[.IndexGroup]
        """
        params = {}
        if name is not None:
            params["name"] = name
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if merge_time is not None:
            params["mergeTime"] = int(merge_time * 1000)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/packet-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_processed_parameter_groups(self):
        """
        Returns the existing parameter groups.

        :rtype: ~collections.abc.Iterable[str]
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/parameter-groups"
        response = self.ctx.get_proto(path=path)
        message = archive_pb2.ParameterGroupInfo()
        message.ParseFromString(response.content)
        groups = getattr(message, "group")
        return iter(groups)

    def list_processed_parameter_group_histogram(
        self, group=None, start=None, stop=None, merge_time=20
    ):
        """
        Reads index records related to processed parameter groups between the
        specified start and stop time.

        Each iteration returns a chunk of chronologically-sorted records.

        :param float merge_time: Maximum gap in seconds before two consecutive index
                                 records are merged together.
        :rtype: ~collections.abc.Iterable[.IndexGroup]
        """
        params = {}
        if group is not None:
            params["group"] = group
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if merge_time is not None:
            params["mergeTime"] = int(merge_time * 1000)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/parameter-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_event_sources(self):
        """
        Returns the existing event sources.

        :rtype: ~collections.abc.Iterable[str]
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/events/sources"
        response = self.ctx.get_proto(path=path)
        message = events_service_pb2.ListEventSourcesResponse()
        message.ParseFromString(response.content)
        sources = getattr(message, "source")
        return iter(sources)

    def list_event_histogram(self, source=None, start=None, stop=None, merge_time=2):
        """
        Reads event-related index records between the specified start and stop
        time.

        Each iteration returns a chunk of chronologically-sorted records.

        :param float merge_time: Maximum gap in seconds before two consecutive index
                                 records are merged together.
        :rtype: ~collections.abc.Iterable[.IndexGroup]
        """
        params = {}
        if source is not None:
            params["source"] = source
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if merge_time is not None:
            params["mergeTime"] = int(merge_time * 1000)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/event-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_command_histogram(self, name=None, start=None, stop=None, merge_time=2):
        """
        Reads command-related index records between the specified start and stop
        time.

        Each iteration returns a chunk of chronologically-sorted records.

        :param float merge_time: Maximum gap in seconds before two consecutive index
                                 records are merged together.
        :rtype: ~collections.abc.Iterable[.IndexGroup]
        """
        params = {}
        if name is not None:
            params["name"] = name
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if merge_time is not None:
            params["mergeTime"] = int(merge_time * 1000)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/command-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_completeness_index(self, start=None, stop=None):
        """
        Reads completeness index records between the specified start and stop
        time.

        Each iteration returns a chunk of chronologically-sorted records.

        :rtype: ~collections.abc.Iterable[.IndexGroup]
        """
        params = {}
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/completeness-index",
            params=params,
            response_class=index_service_pb2.IndexResponse,
            items_key="group",
            item_mapper=IndexGroup,
        )

    def list_packets(
        self, name=None, start=None, stop=None, page_size=500, descending=False
    ):
        """
        Reads packet information between the specified start and stop
        time.

        Packets are sorted by generation time and sequence number.

        :param str name: Archived name of the packet
        :param ~datetime.datetime start: Minimum generation time of the returned
                                         packets (inclusive)
        :param ~datetime.datetime stop: Maximum generation time of the returned
                                        packets (exclusive)
        :param int page_size: Page size of underlying requests. Higher values imply
                              less overhead, but risk hitting the maximum message size
                              limit.
        :param bool descending: If set to ``True`` packets are fetched in reverse
                                order (most recent first).
        :rtype: ~collections.abc.Iterable[.Packet]
        """
        params = {
            "order": "desc" if descending else "asc",
        }
        if name is not None:
            params["name"] = name
        if page_size is not None:
            params["limit"] = page_size
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/packets",
            params=params,
            response_class=packets_service_pb2.ListPacketsResponse,
            items_key="packet",
            item_mapper=Packet,
        )

    def get_packet(self, generation_time, sequence_number):
        """
        Gets a single packet by its identifying key (gentime, seqNum).

        :param ~datetime.datetime generation_time: When the packet was generated
                                                   (packet time)
        :param int sequence_number: Sequence number of the packet
        :rtype: .Packet
        """
        url = f"/archive/{self._instance}/packets/"
        url += f"{to_isostring(generation_time)}/{sequence_number}"
        response = self.ctx.get_proto(url)
        message = packets_pb2.TmPacketData()
        message.ParseFromString(response.content)
        return Packet(message)

    def export_packets(self, name=None, start=None, stop=None, chunk_size=1024):
        """
        Export raw packets.

        Packets are sorted by generation time and sequence number.

        :param str name: Archived name of the packet
        :param ~datetime.datetime start: Minimum generation time of the returned
                                         packets (inclusive)
        :param ~datetime.datetime stop: Maximum generation time of the returned
                                        packets (exclusive)
        :rtype: An iterator over received chunks
        """
        params = {}
        if name is not None:
            params["name"] = name
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        path = f"/archive/{self._instance}:exportPackets"
        response = self.ctx.get_proto(path=path, params=params, stream=True)
        return response.iter_content(chunk_size=chunk_size)

    def list_events(
        self,
        source=None,
        severity=None,
        text_filter=None,
        start=None,
        stop=None,
        page_size=500,
        descending=False,
    ):
        """
        Reads events between the specified start and stop time.

        Events are sorted by generation time, source, then sequence number.

        :param str source: The source of the returned events.
        :param str severity: The minimum severity level of the returned events.
                             One of ``INFO``, ``WATCH``, ``WARNING``, ``DISTRESS``,
                             ``CRITICAL`` or ``SEVERE``.
        :param str text_filter: Filter the text message of the returned events
        :param ~datetime.datetime start: Minimum start date of the returned events
                                         (inclusive)
        :param ~datetime.datetime stop: Maximum start date of the returned events
                                        (exclusive)
        :param int page_size: Page size of underlying requests. Higher values imply
                              less overhead, but risk hitting the maximum message size
                              limit.
        :param bool descending: If set to ``True`` events are fetched in reverse
                                order (most recent first).
        :rtype: ~collections.abc.Iterable[.Event]
        """
        params = {
            "order": "desc" if descending else "asc",
        }
        if source is not None:
            params["source"] = source
        if page_size is not None:
            params["limit"] = page_size
        if severity is not None:
            params["severity"] = severity
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if text_filter is not None:
            params["q"] = text_filter

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/events",
            params=params,
            response_class=events_service_pb2.ListEventsResponse,
            items_key="event",
            item_mapper=Event,
        )

    def sample_parameter_values(
        self,
        parameter,
        start=None,
        stop=None,
        sample_count=500,
        parameter_cache="realtime",
        source="ParameterArchive",
    ):
        """
        Returns parameter samples.

        The query range is split in sample intervals of equal length. For
        each interval a :class:`.Sample` is returned which describes the
        min, max, count and avg during that interval.

        Note that sample times are determined without considering the
        actual parameter values. Two separate queries with equal start/stop
        arguments will always return the same number of samples with the
        same timestamps. This is done to ease merging of multiple sample
        series. You should always be explicit about the ``start`` and ``stop``
        times when relying on this property.

        :param str parameter: Either a fully-qualified XTCE name or an alias in the
                              format ``NAMESPACE/NAME``.
        :param ~datetime.datetime start: Minimum generation time of the sampled
                                         parameter values (inclusive). If not set
                                         this defaults to one hour ago.
        :param ~datetime.datetime stop: Maximum generation time of the sampled
                                        parameter values (exclusive). If not set
                                        this defaults to the current time.
        :param int sample_count: The number of returned samples.
        :param str parameter_cache: Specify the name of the processor who's
                                    parameter cache is merged with already
                                    archived values. To disable results from
                                    the parameter cache, set this to ``None``.
        :param str source: Specify how to retrieve parameter values. By
                           default this uses the ``ParameterArchive`` which
                           is optimized for retrieval. For Yamcs instances
                           that do not enable the ``ParameterArchive``, you can
                           still get results by specifying ``replay`` as the
                           source. Replay requests take longer to return because
                           the data needs to be reprocessed.
        :rtype: .Sample[]
        """
        path = f"/archive/{self._instance}/parameters{parameter}/samples"
        now = datetime.now(tz=timezone.utc)
        params = {
            "count": sample_count,
            "source": source,
            "start": to_isostring(now - timedelta(hours=1)),
            "stop": to_isostring(now),
        }
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        if parameter_cache:
            params["processor"] = parameter_cache
        else:
            params["norealtime"] = True

        response = self.ctx.get_proto(path=path, params=params)
        message = pvalue_pb2.TimeSeries()
        message.ParseFromString(response.content)
        samples = getattr(message, "sample")
        return [Sample(s) for s in samples]

    def list_parameter_ranges(
        self,
        parameter,
        start=None,
        stop=None,
        min_gap=None,
        max_gap=None,
        min_range=None,
        max_values=100,
        parameter_cache="realtime",
    ):
        """
        Returns parameter ranges between the specified start and stop time.

        Each range indicates an interval during which this parameter's
        value was uninterrupted and unchanged.

        Ranges are a good fit for retrieving the value of a parameter
        that does not change frequently. For example an on/off indicator
        or some operational status. Querying ranges will then induce
        much less overhead than manually processing the output of
        :meth:`list_parameter_values` would.

        The maximum number of returned ranges is limited to 500.

        :param str parameter: Either a fully-qualified XTCE name or an alias in the
                              format ``NAMESPACE/NAME``.
        :param start: Minimum generation time of the considered values (inclusive)
        :type start: Optional[~datetime.datetime]
        :param stop: Maximum generation time of the considered values (exclusive)
        :type stop: Optional[~datetime.datetime]
        :param min_gap: Time in seconds. Any gap (detected based on parameter
                        expiration) smaller than this will be ignored.
                        However if the parameter changes value, the ranges
                        will still be split.
        :type max_gap: Optional[float]
        :param max_gap: Time in seconds. If the distance between two
                        subsequent parameter values is bigger than
                        this value (but smaller than the parameter
                        expiration), then an artificial gap is
                        created. This also applies if there is no
                        expiration defined for the parameter.
        :type max_gap: Optional[float]
        :param min_range: Time in seconds. Minimum duration of returned
                          ranges. If multiple values occur within the
                          range, the most frequent can be accessed using
                          the ``entries`` property.
        :type min_range: Optional[float]
        :param max_values: Maximum number of unique values, tallied across the
                           full requested range. Use this in combination with
                           ``min_range`` to further optimize for transfer size.
                           This value is limited to 100 at most.
        :param str parameter_cache: Specify the name of the processor who's
                                    parameter cache is merged with already
                                    archived values. To disable results from
                                    the parameter cache, set this to ``None``.
        :type parameter_cache: Optional[str]
        :rtype: .ParameterRange[]
        """
        path = f"/archive/{self._instance}/parameters{parameter}/ranges"
        params = {}
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)
        if min_gap is not None:
            params["minGap"] = int(min_gap * 1000)
        if max_gap is not None:
            params["maxGap"] = int(max_gap * 1000)
        if min_range is not None:
            params["minRange"] = int(min_range * 1000)
        if max_values is not None:
            params["maxValues"] = max_values

        if parameter_cache:
            params["processor"] = parameter_cache
        else:
            params["norealtime"] = True

        response = self.ctx.get_proto(path=path, params=params)
        message = pvalue_pb2.Ranges()
        message.ParseFromString(response.content)
        ranges = getattr(message, "range")
        return [ParameterRange(r) for r in ranges]

    def list_parameter_values(
        self,
        parameter,
        start=None,
        stop=None,
        page_size=500,
        descending=False,
        parameter_cache="realtime",
        source="ParameterArchive",
    ):
        """
        Reads parameter values between the specified start and stop time.

        .. note::

            This method will send out multiple requests when more than
            ``page_size`` values are queried. For large queries, consider
            using :meth:`stream_parameter_values` instead, it uses server-streaming
            based on a single request, and supports downloading the values of
            multiple parameter at the same time.

        :param str parameter: Either a fully-qualified XTCE name or an alias in the
                              format ``NAMESPACE/NAME``.
        :param ~datetime.datetime start: Minimum generation time of the returned
                                         values (inclusive)
        :param ~datetime.datetime stop: Maximum generation time of the returned
                                        values (exclusive)
        :param int page_size: Page size of underlying requests. Higher values imply
                              less overhead, but risk hitting the maximum message size
                              limit.
        :param bool descending: If set to ``True`` values are fetched in reverse
                                order (most recent first).
        :param str parameter_cache: Specify the name of the processor who's
                                    parameter cache is merged with already
                                    archived values. To disable results from
                                    the parameter cache, set this to ``None``.
        :param str source: Specify how to retrieve parameter values. By
                           default this uses the ``ParameterArchive`` which
                           is optimized for retrieval. For Yamcs instances
                           that do not enable the ``ParameterArchive``, you can
                           still get results by specifying ``replay`` as the
                           source. Replay requests take longer to return because
                           the data needs to be reprocessed.
        :rtype: ~collections.abc.Iterable[.ParameterValue]
        """
        params = {
            "source": source,
            "order": "desc" if descending else "asc",
        }
        if page_size is not None:
            params["limit"] = page_size
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        if parameter_cache:
            params["processor"] = parameter_cache
        else:
            params["norealtime"] = True

        return pagination.Iterator(
            ctx=self.ctx,
            path=f"/archive/{self._instance}/parameters{parameter}",
            params=params,
            response_class=archive_pb2.ListParameterHistoryResponse,
            items_key="parameter",
            item_mapper=ParameterValue,
        )

    def stream_parameter_values(
        self, parameters, start=None, stop=None, chunk_size=32 * 1024
    ):
        """
        Reads parameter values between the specified start and stop time.

        Value updates are emitted for each unique generation time within
        the queried range. If one of the parameters does not have a value
        for a specific generation time, it is not included in the update.

        :param parameters: Parameter(s) to be queried.
        :type parameters: Union[str, str[]]
        :param ~datetime.datetime start: Minimum generation time of the returned
                                         values (inclusive)
        :param ~datetime.datetime stop: Maximum generation time of the returned
                                        values (exclusive)
        :return: :rtype: ~collections.abc.Iterable[:class:`.ParameterData`]
        """
        options = archive_pb2.StreamParameterValuesRequest()
        options.ids.extend(to_named_object_ids(parameters))
        if start is not None:
            options.start.MergeFrom(to_server_time(start))
        if stop is not None:
            options.stop.MergeFrom(to_server_time(stop))

        def generate():
            path = f"/stream-archive/{self._instance}:streamParameterValues"
            response = self.ctx.post_proto(
                path=path, data=options.SerializeToString(), stream=True
            )
            for message in split_protobuf_stream(
                response.iter_content(chunk_size=chunk_size), pvalue_pb2.ParameterData
            ):
                yield ParameterData(message)

        return generate()

    def list_command_history(
        self,
        command=None,
        queue=None,
        start=None,
        stop=None,
        page_size=500,
        descending=False,
    ):
        """
        Reads command history entries between the specified start and stop time.

        :param str command: Either a fully-qualified XTCE name or an alias in the
                            format ``NAMESPACE/NAME``.
        :param str queue: Name of the queue that the command was assigned to.
        :param ~datetime.datetime start: Minimum generation time of the returned
                                         command history entries (inclusive)
        :param ~datetime.datetime stop: Maximum generation time of the returned
                                        command history entries (exclusive)
        :param int page_size: Page size of underlying requests. Higher values imply
                              less overhead, but risk hitting the maximum message size
                              limit.
        :param bool descending: If set to ``True`` results are fetched in reverse
                                order (most recent first).
        :rtype: ~collections.abc.Iterable[.CommandHistory]
        """
        params = {
            "order": "desc" if descending else "asc",
        }
        if queue:
            params["queue"] = queue
        if page_size is not None:
            params["limit"] = page_size
        if start is not None:
            params["start"] = to_isostring(start)
        if stop is not None:
            params["stop"] = to_isostring(stop)

        if command:
            path = f"/archive/{self._instance}/commands{command}"
        else:
            path = f"/archive/{self._instance}/commands"

        return pagination.Iterator(
            ctx=self.ctx,
            path=path,
            params=params,
            response_class=commands_service_pb2.ListCommandsResponse,
            items_key="entry",
            item_mapper=CommandHistory,
        )

    def list_tables(self):
        """
        Returns the existing tables.

        Tables are returned in lexicographical order.

        :rtype: ~collections.abc.Iterable[.Table]
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/tables"
        response = self.ctx.get_proto(path=path)
        message = table_pb2.ListTablesResponse()
        message.ParseFromString(response.content)
        tables = getattr(message, "tables")
        return iter([Table(table) for table in tables])

    def get_table(self, table):
        """
        Gets a single table.

        :param str table: The name of the table.
        :rtype: .Table
        """
        path = f"/archive/{self._instance}/tables/{table}"
        response = self.ctx.get_proto(path=path)
        message = table_pb2.TableInfo()
        message.ParseFromString(response.content)
        return Table(message)

    def dump_table(self, table, chunk_size=32 * 1024):
        path = f"/archive/{self._instance}/tables/{table}:readRows"
        response = self.ctx.post_proto(path=path, stream=True)
        return response.iter_content(chunk_size=chunk_size)

    def load_table(self, table, data, chunk_size=32 * 1024):
        def read_in_chunks(file_object, chunk_size):
            chunk = file_object.read(chunk_size)
            while chunk:
                yield chunk
                chunk = file_object.read(chunk_size)

        path = f"/archive/{self._instance}/tables/{table}:writeRows"
        if chunk_size:
            generator = read_in_chunks(data, chunk_size)
            response = self.ctx.post_proto(path=path, data=generator)
        else:
            response = self.ctx.post_proto(path=path, data=data)
        message = table_pb2.WriteRowsResponse()
        message.ParseFromString(response.content)
        return message.count

    def rebuild_histogram(self, table, start=None, stop=None):
        """
        Rebuilds the histogram for a table. This may be necessary
        for example after bulk loading data.

        The rebuild may be constrained by using the
        ``start`` and ``stop`` parameters. When
        specified all partitions overlapping this range
        are reconsidered.

        .. note::
            Histogram rebuilds run synchronously: this
            method will await the outcome.

        :param str table: The name of the table
        :param start: Start time
        :type start: Optional[~datetime.datetime]
        :param stop: Stop time
        :type stop: Optional[~datetime.datetime]
        """
        req = table_pb2.RebuildHistogramRequest()
        if start:
            req.start.MergeFrom(to_server_time(start))
        if stop:
            req.stop.MergeFrom(to_server_time(stop))
        url = f"/archive/{self._instance}/tables/{table}:rebuildHistogram"
        self.ctx.post_proto(url, data=req.SerializeToString())

    def rebuild_parameter_archive(self, start, stop):
        """
        Rebuilds the Parameter Archive.

        The rebuild must be constrained by using the
        ``start`` and ``stop`` parameters. This values
        are only hints to the Parameter Archive, which
        will extend the requested range based on archive
        segmentation.

        .. note::
            Rebuilds run as an asynchronous operation: this
            method will not await the outcome.

        :param start: Start time
        :type start: ~datetime.datetime
        :param stop: Stop time
        :type stop: ~datetime.datetime
        """
        req = parameter_archive_service_pb2.RebuildRangeRequest()
        req.start.MergeFrom(to_server_time(start))
        req.stop.MergeFrom(to_server_time(stop))
        url = f"/archive/{self._instance}/parameterArchive:rebuild"
        self.ctx.post_proto(url, data=req.SerializeToString())

    def list_streams(self):
        """
        Returns the existing streams.

        Streams are returned in lexicographical order.

        :rtype: ~collections.abc.Iterable[.Stream]
        """
        # Server does not do pagination on listings of this resource.
        # Return an iterator anyway for similarity with other API methods
        path = f"/archive/{self._instance}/streams"
        response = self.ctx.get_proto(path=path)
        message = table_pb2.ListStreamsResponse()
        message.ParseFromString(response.content)
        streams = getattr(message, "streams")
        return iter([Stream(stream) for stream in streams])

    def get_stream(self, stream):
        """
        Gets a single stream.

        :param str stream: The name of the stream.
        :rtype: .Stream
        """
        path = f"/archive/{self._instance}/streams/{stream}"
        response = self.ctx.get_proto(path=path)
        message = table_pb2.StreamInfo()
        message.ParseFromString(response.content)
        return Stream(message)

    def create_stream_subscription(self, stream, on_data, timeout=60):
        """
        Create a new stream subscription.

        :param str stream: The name of the stream.
        :param on_data: Function that gets called with  :class:`.StreamData`
                        updates.
        :param float timeout: The amount of seconds to wait for the request
                              to complete.
        :return: Future that can be used to manage the background websocket
                 subscription
        :rtype: .WebSocketSubscriptionFuture
        """
        options = table_pb2.SubscribeStreamRequest()
        options.instance = self._instance
        options.stream = stream

        manager = WebSocketSubscriptionManager(
            self.ctx, topic="stream", options=options
        )

        # Represent subscription as a future
        subscription = WebSocketSubscriptionFuture(manager)

        wrapped_callback = functools.partial(
            _wrap_callback_parse_stream_data, subscription, on_data
        )

        manager.open(wrapped_callback)

        # Wait until a reply or exception is received
        subscription.reply(timeout=timeout)

        return subscription

    def execute_sql(self, statement):
        """
        Executes a single SQL statement.

        :param statement: SQL string
        :return: A result set for consuming rows
        :rtype: .ResultSet
        """
        path = f"/archive/{self._instance}:executeStreamingSql"
        req = table_pb2.ExecuteSqlRequest()
        req.statement = statement

        response = self.ctx.post_proto(
            path=path, data=req.SerializeToString(), stream=True
        )
        return ResultSet(response)



from datetime import datetime, timedelta, timezone

from requests.auth import HTTPBasicAuth
from yamcs.core.exceptions import Unauthorized, YamcsError


def _convert_user_credentials(
    session, token_url, username=None, password=None, refresh_token=None
):
    """
    Converts username/password credentials to token credentials by
    using Yamcs as the authentication server.
    """
    if username and password:
        data = {"grant_type": "password", "username": username, "password": password}
    elif refresh_token:
        data = {"grant_type": "refresh_token", "refresh_token": refresh_token}
    else:
        raise NotImplementedError()

    response = session.post(token_url, data=data)
    if response.status_code == 401:
        raise Unauthorized("401 Client Error: Unauthorized")
    elif response.status_code == 200:
        d = response.json()
        expiry = datetime.now(tz=timezone.utc) + timedelta(seconds=d["expires_in"])
        return Credentials(
            access_token=d["access_token"],
            refresh_token=d["refresh_token"],
            expiry=expiry,
        )
    else:
        raise YamcsError(f"{response.status_code} Server Error")


def _convert_service_account_credentials(
    session, token_url, client_id, client_secret, become
):
    """
    Converts service account credentials to impersonated token credentials.
    """
    data = {"grant_type": "client_credentials", "become": become}

    response = session.post(
        token_url, data=data, auth=HTTPBasicAuth(client_id, client_secret)
    )
    if response.status_code == 401:
        raise Unauthorized("401 Client Error: Unauthorized")
    elif response.status_code == 200:
        d = response.json()
        expiry = datetime.now(tz=timezone.utc) + timedelta(seconds=d["expires_in"])
        return Credentials(
            access_token=d["access_token"],
            client_id=client_id,
            client_secret=client_secret,
            become=become,
            expiry=expiry,
        )
    else:
        raise YamcsError(f"{response.status_code} Server Error")


class Credentials:
    """
    Data holder for different types of credentials. Currently this includes:

    * Username/password credentials (fields ``username`` and ``password``)
    * Bearer tokens (fields ``access_token`` and optionally ``refresh_token``)
    """

    def __init__(
        self,
        username=None,
        password=None,
        access_token=None,
        refresh_token=None,
        expiry=None,
        client_id=None,
        client_secret=None,
        become=None,
    ):
        self.username = username
        """Username (only needed when using username/password credentials)."""

        self.password = password
        """Clear-text password (consider TLS!)."""

        self.access_token = access_token
        """Short-lived bearer token."""

        self.refresh_token = refresh_token
        """Refresh token used to request a new access token."""

        self.expiry = expiry
        """When this token expires."""

        self.client_id = client_id
        """The client ID. Currently used only by service accounts."""

        self.client_secret = client_secret
        """The client secret. Currently used only by service accounts."""

        self.become = become
        """
        Name of the user to impersonate. Only service accounts with
        impersonation authority can use this feature.
        """

        self._on_token_update = None

    def login(self, session, auth_url, on_token_update):
        self._on_token_update = on_token_update
        token_url = auth_url + "/token"
        if self.username:  # Convert u/p to bearer
            creds = _convert_user_credentials(
                session, token_url, username=self.username, password=self.password
            )
        elif self.become:  # Impersonate from service account
            creds = _convert_service_account_credentials(
                session,
                token_url,
                client_id=self.client_id,
                client_secret=self.client_secret,
                become=self.become,
            )
        else:
            creds = self

        if on_token_update:
            on_token_update(creds)
        return creds

    def is_expired(self):
        return self.expiry and datetime.now(tz=timezone.utc) >= self.expiry

    def refresh(self, session, auth_url):
        if self.become:
            new_creds = _convert_service_account_credentials(
                session,
                auth_url + "/token",
                client_id=self.client_id,
                client_secret=self.client_secret,
                become=self.become,
            )
        elif self.refresh_token:
            new_creds = _convert_user_credentials(
                session, auth_url + "/token", refresh_token=self.refresh_token
            )
        else:
            raise YamcsError("Missing refresh token")

        self.access_token = new_creds.access_token
        self.refresh_token = new_creds.refresh_token
        self.expiry = new_creds.expiry
        if self._on_token_update:
            self._on_token_update(self)

    def before_request(self, session, auth_url):
        if self.is_expired():
            self.refresh(session, auth_url)

        if self.access_token:
            session.headers.update({"Authorization": "Bearer " + self.access_token})



import requests
import urllib3
from google.protobuf.message import DecodeError
from yamcs import clientversion
from yamcs.api import exception_pb2
from yamcs.core.exceptions import ConnectionFailure, NotFound, Unauthorized, YamcsError


class Context:

    credentials = None

    def __init__(
        self,
        address,
        tls=False,
        credentials=None,
        user_agent=None,
        on_token_update=None,
        tls_verify=True,
    ):
        if ":" in address:
            self.address = address
        else:
            self.address = address + ":8090"

        if tls:
            self.url = f"https://{self.address}"
            self.auth_root = f"https://{self.address}/auth"
            self.api_root = f"https://{self.address}/api"
            self.ws_root = f"wss://{self.address}/api/websocket"
        else:
            self.url = f"http://{self.address}"
            self.auth_root = f"http://{self.address}/auth"
            self.api_root = f"http://{self.address}/api"
            self.ws_root = f"ws://{self.address}/api/websocket"

        self.session = requests.Session()
        self.session.verify = tls_verify
        if not tls_verify:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        if credentials:
            self.credentials = credentials.login(
                self.session, self.auth_root, on_token_update
            )

        if not user_agent:
            user_agent = "python-yamcs-client v" + clientversion.__version__
        self.session.headers.update({"User-Agent": user_agent})

    def get_proto(self, path, **kwargs):
        headers = kwargs.pop("headers", {})
        headers["Accept"] = "application/protobuf"
        kwargs.update({"headers": headers})
        return self.request("get", path, **kwargs)

    def put_proto(self, path, **kwargs):
        headers = kwargs.pop("headers", {})
        headers["Content-Type"] = "application/protobuf"
        headers["Accept"] = "application/protobuf"
        kwargs.update({"headers": headers})
        return self.request("put", path, **kwargs)

    def patch_proto(self, path, **kwargs):
        headers = kwargs.pop("headers", {})
        headers["Content-Type"] = "application/protobuf"
        headers["Accept"] = "application/protobuf"
        kwargs.update({"headers": headers})
        return self.request("patch", path, **kwargs)

    def post_proto(self, path, **kwargs):
        headers = kwargs.pop("headers", {})
        headers["Content-Type"] = "application/protobuf"
        headers["Accept"] = "application/protobuf"
        kwargs.update({"headers": headers})
        return self.request("post", path, **kwargs)

    def delete_proto(self, path, **kwargs):
        headers = kwargs.pop("headers", {})
        headers["Accept"] = "application/protobuf"
        kwargs.update({"headers": headers})
        return self.request("delete", path, **kwargs)

    def request(self, method, path, **kwargs):
        path = f"{self.api_root}{path}"

        try:
            if self.credentials:
                self.credentials.before_request(self.session, self.auth_root)

            response = self.session.request(method, path, **kwargs)
        except requests.exceptions.SSLError as ssl_error:
            msg = f"Connection to {self.url} failed: {ssl_error}"
            raise ConnectionFailure(msg) from None
        except requests.exceptions.ConnectionError as e:
            # Requests gives us a horribly confusing error when a connection
            # is refused. Confirm and unwrap.
            if e.args and isinstance(e.args[0], urllib3.exceptions.MaxRetryError):
                # This is a string (which is still confusing ....)
                msg = e.args[0].args[0]
                if "refused" in msg:
                    msg = f"Connection to {self.url} failed: connection refused"
                    raise ConnectionFailure(msg) from None
                elif "not known" in msg:
                    msg = f"Connection to {self.url} failed: could not resolve hostname"
                    raise ConnectionFailure(msg) from None

            raise ConnectionFailure(f"Connection to {self.url} failed: {e}")

        if 200 <= response.status_code < 300:
            return response

        exception_message = exception_pb2.ExceptionMessage()
        try:
            exception_message.ParseFromString(response.content)
        except DecodeError:
            pass

        if response.status_code == 401:
            raise Unauthorized("401 Client Error: Unauthorized")
        elif response.status_code == 404:
            msg = getattr(exception_message, "msg")
            raise NotFound(f"404 Client Error: {msg}")
        elif 400 <= response.status_code < 500:
            msg = getattr(exception_message, "msg")
            raise YamcsError(f"{response.status_code} Client Error: {msg}")
        msg = getattr(exception_message, "msg")
        raise YamcsError(f"{response.status_code} Server Error: {msg}")



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yamcs/protobuf/web/server_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from yamcs.api import annotations_pb2 as yamcs_dot_api_dot_annotations__pb2
from yamcs.api import httpbody_pb2 as yamcs_dot_api_dot_httpbody__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='yamcs/protobuf/web/server_service.proto',
  package='yamcs.protobuf.web',
  syntax='proto2',
  serialized_options=_b('\n\022org.yamcs.protobufB\022ServerServiceProtoP\001'),
  serialized_pb=_b('\n\'yamcs/protobuf/web/server_service.proto\x12\x12yamcs.protobuf.web\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1byamcs/api/annotations.proto\x1a\x18yamcs/api/httpbody.proto\"\x1e\n\x10GetThreadRequest\x12\n\n\x02id\x18\x01 \x01(\x03\"F\n\x13ListThreadsResponse\x12/\n\x07threads\x18\x01 \x03(\x0b\x32\x1e.yamcs.protobuf.web.ThreadInfo\"C\n\x12ListRoutesResponse\x12-\n\x06routes\x18\x01 \x03(\x0b\x32\x1d.yamcs.protobuf.web.RouteInfo\"C\n\x12ListTopicsResponse\x12-\n\x06topics\x18\x01 \x03(\x0b\x32\x1d.yamcs.protobuf.web.TopicInfo\"\xa6\x01\n\x0bHttpTraffic\x12\x11\n\treadBytes\x18\x01 \x01(\x04\x12\x14\n\x0cwrittenBytes\x18\x02 \x01(\x04\x12\x16\n\x0ereadThroughput\x18\x03 \x01(\x04\x12\x17\n\x0fwriteThroughput\x18\x04 \x01(\x04\x12=\n\x0b\x63onnections\x18\x05 \x03(\x0b\x32(.yamcs.protobuf.web.ClientConnectionInfo\"\xda\x01\n\tRouteInfo\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x12\n\nhttpMethod\x18\x04 \x01(\t\x12\x0b\n\x03url\x18\x05 \x01(\t\x12\x11\n\tinputType\x18\x06 \x01(\t\x12\x12\n\noutputType\x18\x07 \x01(\t\x12\x12\n\ndeprecated\x18\x08 \x01(\x08\x12\x14\n\x0crequestCount\x18\t \x01(\x03\x12\x12\n\nerrorCount\x18\n \x01(\x03\x12\x11\n\tlogFormat\x18\x0b \x01(\t\"\x8b\x01\n\tTopicInfo\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x0f\n\x07service\x18\x02 \x01(\t\x12\x0e\n\x06method\x18\x03 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x11\n\tinputType\x18\x05 \x01(\t\x12\x12\n\noutputType\x18\x06 \x01(\t\x12\x12\n\ndeprecated\x18\x07 \x01(\x08\"_\n\x10TraceElementInfo\x12\x11\n\tclassName\x18\x01 \x01(\t\x12\x10\n\x08\x66ileName\x18\x02 \x01(\t\x12\x12\n\nmethodName\x18\x03 \x01(\t\x12\x12\n\nlineNumber\x18\x04 \x01(\x05\"#\n\x12ListThreadsRequest\x12\r\n\x05\x64\x65pth\x18\x01 \x01(\x05\"d\n\x0fThreadGroupInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06\x64\x61\x65mon\x18\x02 \x01(\x08\x12\x33\n\x06parent\x18\x03 \x01(\x0b\x32#.yamcs.protobuf.web.ThreadGroupInfo\"\xc1\x01\n\nThreadInfo\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05state\x18\x03 \x01(\t\x12\x0e\n\x06native\x18\x04 \x01(\x08\x12\x11\n\tsuspended\x18\x05 \x01(\x08\x12\x32\n\x05group\x18\x06 \x01(\x0b\x32#.yamcs.protobuf.web.ThreadGroupInfo\x12\x33\n\x05trace\x18\x07 \x03(\x0b\x32$.yamcs.protobuf.web.TraceElementInfo\"\xaf\x03\n\x15GetServerInfoResponse\x12\x14\n\x0cyamcsVersion\x18\x01 \x01(\t\x12\x10\n\x08revision\x18\x06 \x01(\t\x12\x10\n\x08serverId\x18\x02 \x01(\t\x12\x1c\n\x14\x64\x65\x66\x61ultYamcsInstance\x18\x03 \x01(\t\x12\x45\n\x07plugins\x18\x05 \x03(\x0b\x32\x34.yamcs.protobuf.web.GetServerInfoResponse.PluginInfo\x12S\n\x0e\x63ommandOptions\x18\x07 \x03(\x0b\x32;.yamcs.protobuf.web.GetServerInfoResponse.CommandOptionInfo\x1aP\n\nPluginInfo\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x0e\n\x06vendor\x18\x04 \x01(\t\x1aP\n\x11\x43ommandOptionInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12\x13\n\x0bverboseName\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\x0c\n\x04help\x18\x04 \x01(\t\"\x8c\x03\n\x14\x43lientConnectionInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04open\x18\x02 \x01(\x08\x12\x0e\n\x06\x61\x63tive\x18\x03 \x01(\x08\x12\x10\n\x08writable\x18\x04 \x01(\x08\x12\x15\n\rremoteAddress\x18\x05 \x01(\t\x12\x11\n\treadBytes\x18\x06 \x01(\x04\x12\x14\n\x0cwrittenBytes\x18\x07 \x01(\x04\x12\x16\n\x0ereadThroughput\x18\x08 \x01(\x04\x12\x17\n\x0fwriteThroughput\x18\t \x01(\x04\x12M\n\x0bhttpRequest\x18\n \x01(\x0b\x32\x38.yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo\x12\x10\n\x08username\x18\x0b \x01(\t\x1a\x66\n\x0fHttpRequestInfo\x12\x10\n\x08protocol\x18\x01 \x01(\t\x12\x0e\n\x06method\x18\x02 \x01(\t\x12\x0b\n\x03uri\x18\x03 \x01(\t\x12\x11\n\tkeepAlive\x18\x04 \x01(\x08\x12\x11\n\tuserAgent\x18\x05 \x01(\t2\xc6\x05\n\tServerApi\x12^\n\rGetServerInfo\x12\x16.google.protobuf.Empty\x1a).yamcs.protobuf.web.GetServerInfoResponse\"\n\x8a\x92\x03\x06\n\x04/api\x12_\n\nListRoutes\x12\x16.google.protobuf.Empty\x1a&.yamcs.protobuf.web.ListRoutesResponse\"\x11\x8a\x92\x03\r\n\x0b/api/routes\x12_\n\nListTopics\x12\x16.google.protobuf.Empty\x1a&.yamcs.protobuf.web.ListTopicsResponse\"\x11\x8a\x92\x03\r\n\x0b/api/topics\x12r\n\x0bListThreads\x12&.yamcs.protobuf.web.ListThreadsRequest\x1a\'.yamcs.protobuf.web.ListThreadsResponse\"\x12\x8a\x92\x03\x0e\n\x0c/api/threads\x12j\n\tGetThread\x12$.yamcs.protobuf.web.GetThreadRequest\x1a\x1e.yamcs.protobuf.web.ThreadInfo\"\x17\x8a\x92\x03\x13\n\x11/api/threads/{id}\x12S\n\x0b\x44umpThreads\x12\x16.google.protobuf.Empty\x1a\x13.yamcs.api.HttpBody\"\x17\x8a\x92\x03\x13\n\x11/api/threads:dump\x12\x62\n\x0eGetHttpTraffic\x12\x16.google.protobuf.Empty\x1a\x1f.yamcs.protobuf.web.HttpTraffic\"\x17\x8a\x92\x03\x13\n\x11/api/http-trafficB*\n\x12org.yamcs.protobufB\x12ServerServiceProtoP\x01')
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,yamcs_dot_api_dot_annotations__pb2.DESCRIPTOR,yamcs_dot_api_dot_httpbody__pb2.DESCRIPTOR,])




_GETTHREADREQUEST = _descriptor.Descriptor(
  name='GetThreadRequest',
  full_name='yamcs.protobuf.web.GetThreadRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='yamcs.protobuf.web.GetThreadRequest.id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=147,
  serialized_end=177,
)


_LISTTHREADSRESPONSE = _descriptor.Descriptor(
  name='ListThreadsResponse',
  full_name='yamcs.protobuf.web.ListThreadsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='threads', full_name='yamcs.protobuf.web.ListThreadsResponse.threads', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=179,
  serialized_end=249,
)


_LISTROUTESRESPONSE = _descriptor.Descriptor(
  name='ListRoutesResponse',
  full_name='yamcs.protobuf.web.ListRoutesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='routes', full_name='yamcs.protobuf.web.ListRoutesResponse.routes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=251,
  serialized_end=318,
)


_LISTTOPICSRESPONSE = _descriptor.Descriptor(
  name='ListTopicsResponse',
  full_name='yamcs.protobuf.web.ListTopicsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='topics', full_name='yamcs.protobuf.web.ListTopicsResponse.topics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=320,
  serialized_end=387,
)


_HTTPTRAFFIC = _descriptor.Descriptor(
  name='HttpTraffic',
  full_name='yamcs.protobuf.web.HttpTraffic',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='readBytes', full_name='yamcs.protobuf.web.HttpTraffic.readBytes', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='writtenBytes', full_name='yamcs.protobuf.web.HttpTraffic.writtenBytes', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='readThroughput', full_name='yamcs.protobuf.web.HttpTraffic.readThroughput', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='writeThroughput', full_name='yamcs.protobuf.web.HttpTraffic.writeThroughput', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='connections', full_name='yamcs.protobuf.web.HttpTraffic.connections', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=390,
  serialized_end=556,
)


_ROUTEINFO = _descriptor.Descriptor(
  name='RouteInfo',
  full_name='yamcs.protobuf.web.RouteInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='service', full_name='yamcs.protobuf.web.RouteInfo.service', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='method', full_name='yamcs.protobuf.web.RouteInfo.method', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='yamcs.protobuf.web.RouteInfo.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='httpMethod', full_name='yamcs.protobuf.web.RouteInfo.httpMethod', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='url', full_name='yamcs.protobuf.web.RouteInfo.url', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inputType', full_name='yamcs.protobuf.web.RouteInfo.inputType', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='outputType', full_name='yamcs.protobuf.web.RouteInfo.outputType', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deprecated', full_name='yamcs.protobuf.web.RouteInfo.deprecated', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='requestCount', full_name='yamcs.protobuf.web.RouteInfo.requestCount', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='errorCount', full_name='yamcs.protobuf.web.RouteInfo.errorCount', index=9,
      number=10, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='logFormat', full_name='yamcs.protobuf.web.RouteInfo.logFormat', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=559,
  serialized_end=777,
)


_TOPICINFO = _descriptor.Descriptor(
  name='TopicInfo',
  full_name='yamcs.protobuf.web.TopicInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='topic', full_name='yamcs.protobuf.web.TopicInfo.topic', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='service', full_name='yamcs.protobuf.web.TopicInfo.service', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='method', full_name='yamcs.protobuf.web.TopicInfo.method', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='yamcs.protobuf.web.TopicInfo.description', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inputType', full_name='yamcs.protobuf.web.TopicInfo.inputType', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='outputType', full_name='yamcs.protobuf.web.TopicInfo.outputType', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deprecated', full_name='yamcs.protobuf.web.TopicInfo.deprecated', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=780,
  serialized_end=919,
)


_TRACEELEMENTINFO = _descriptor.Descriptor(
  name='TraceElementInfo',
  full_name='yamcs.protobuf.web.TraceElementInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='className', full_name='yamcs.protobuf.web.TraceElementInfo.className', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='fileName', full_name='yamcs.protobuf.web.TraceElementInfo.fileName', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='methodName', full_name='yamcs.protobuf.web.TraceElementInfo.methodName', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lineNumber', full_name='yamcs.protobuf.web.TraceElementInfo.lineNumber', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=921,
  serialized_end=1016,
)


_LISTTHREADSREQUEST = _descriptor.Descriptor(
  name='ListThreadsRequest',
  full_name='yamcs.protobuf.web.ListThreadsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='depth', full_name='yamcs.protobuf.web.ListThreadsRequest.depth', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1018,
  serialized_end=1053,
)


_THREADGROUPINFO = _descriptor.Descriptor(
  name='ThreadGroupInfo',
  full_name='yamcs.protobuf.web.ThreadGroupInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='yamcs.protobuf.web.ThreadGroupInfo.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='daemon', full_name='yamcs.protobuf.web.ThreadGroupInfo.daemon', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parent', full_name='yamcs.protobuf.web.ThreadGroupInfo.parent', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1055,
  serialized_end=1155,
)


_THREADINFO = _descriptor.Descriptor(
  name='ThreadInfo',
  full_name='yamcs.protobuf.web.ThreadInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='yamcs.protobuf.web.ThreadInfo.id', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='yamcs.protobuf.web.ThreadInfo.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='state', full_name='yamcs.protobuf.web.ThreadInfo.state', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='native', full_name='yamcs.protobuf.web.ThreadInfo.native', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='suspended', full_name='yamcs.protobuf.web.ThreadInfo.suspended', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='group', full_name='yamcs.protobuf.web.ThreadInfo.group', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace', full_name='yamcs.protobuf.web.ThreadInfo.trace', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1158,
  serialized_end=1351,
)


_GETSERVERINFORESPONSE_PLUGININFO = _descriptor.Descriptor(
  name='PluginInfo',
  full_name='yamcs.protobuf.web.GetServerInfoResponse.PluginInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='yamcs.protobuf.web.GetServerInfoResponse.PluginInfo.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='yamcs.protobuf.web.GetServerInfoResponse.PluginInfo.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='yamcs.protobuf.web.GetServerInfoResponse.PluginInfo.version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vendor', full_name='yamcs.protobuf.web.GetServerInfoResponse.PluginInfo.vendor', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1623,
  serialized_end=1703,
)

_GETSERVERINFORESPONSE_COMMANDOPTIONINFO = _descriptor.Descriptor(
  name='CommandOptionInfo',
  full_name='yamcs.protobuf.web.GetServerInfoResponse.CommandOptionInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='yamcs.protobuf.web.GetServerInfoResponse.CommandOptionInfo.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='verboseName', full_name='yamcs.protobuf.web.GetServerInfoResponse.CommandOptionInfo.verboseName', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='yamcs.protobuf.web.GetServerInfoResponse.CommandOptionInfo.type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='help', full_name='yamcs.protobuf.web.GetServerInfoResponse.CommandOptionInfo.help', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1705,
  serialized_end=1785,
)

_GETSERVERINFORESPONSE = _descriptor.Descriptor(
  name='GetServerInfoResponse',
  full_name='yamcs.protobuf.web.GetServerInfoResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='yamcsVersion', full_name='yamcs.protobuf.web.GetServerInfoResponse.yamcsVersion', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='revision', full_name='yamcs.protobuf.web.GetServerInfoResponse.revision', index=1,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='serverId', full_name='yamcs.protobuf.web.GetServerInfoResponse.serverId', index=2,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='defaultYamcsInstance', full_name='yamcs.protobuf.web.GetServerInfoResponse.defaultYamcsInstance', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='plugins', full_name='yamcs.protobuf.web.GetServerInfoResponse.plugins', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='commandOptions', full_name='yamcs.protobuf.web.GetServerInfoResponse.commandOptions', index=5,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_GETSERVERINFORESPONSE_PLUGININFO, _GETSERVERINFORESPONSE_COMMANDOPTIONINFO, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1354,
  serialized_end=1785,
)


_CLIENTCONNECTIONINFO_HTTPREQUESTINFO = _descriptor.Descriptor(
  name='HttpRequestInfo',
  full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='protocol', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.protocol', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='method', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.method', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uri', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.uri', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keepAlive', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.keepAlive', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='userAgent', full_name='yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo.userAgent', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2082,
  serialized_end=2184,
)

_CLIENTCONNECTIONINFO = _descriptor.Descriptor(
  name='ClientConnectionInfo',
  full_name='yamcs.protobuf.web.ClientConnectionInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='yamcs.protobuf.web.ClientConnectionInfo.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='open', full_name='yamcs.protobuf.web.ClientConnectionInfo.open', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='active', full_name='yamcs.protobuf.web.ClientConnectionInfo.active', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='writable', full_name='yamcs.protobuf.web.ClientConnectionInfo.writable', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='remoteAddress', full_name='yamcs.protobuf.web.ClientConnectionInfo.remoteAddress', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='readBytes', full_name='yamcs.protobuf.web.ClientConnectionInfo.readBytes', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='writtenBytes', full_name='yamcs.protobuf.web.ClientConnectionInfo.writtenBytes', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='readThroughput', full_name='yamcs.protobuf.web.ClientConnectionInfo.readThroughput', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='writeThroughput', full_name='yamcs.protobuf.web.ClientConnectionInfo.writeThroughput', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='httpRequest', full_name='yamcs.protobuf.web.ClientConnectionInfo.httpRequest', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='username', full_name='yamcs.protobuf.web.ClientConnectionInfo.username', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_CLIENTCONNECTIONINFO_HTTPREQUESTINFO, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1788,
  serialized_end=2184,
)

_LISTTHREADSRESPONSE.fields_by_name['threads'].message_type = _THREADINFO
_LISTROUTESRESPONSE.fields_by_name['routes'].message_type = _ROUTEINFO
_LISTTOPICSRESPONSE.fields_by_name['topics'].message_type = _TOPICINFO
_HTTPTRAFFIC.fields_by_name['connections'].message_type = _CLIENTCONNECTIONINFO
_THREADGROUPINFO.fields_by_name['parent'].message_type = _THREADGROUPINFO
_THREADINFO.fields_by_name['group'].message_type = _THREADGROUPINFO
_THREADINFO.fields_by_name['trace'].message_type = _TRACEELEMENTINFO
_GETSERVERINFORESPONSE_PLUGININFO.containing_type = _GETSERVERINFORESPONSE
_GETSERVERINFORESPONSE_COMMANDOPTIONINFO.containing_type = _GETSERVERINFORESPONSE
_GETSERVERINFORESPONSE.fields_by_name['plugins'].message_type = _GETSERVERINFORESPONSE_PLUGININFO
_GETSERVERINFORESPONSE.fields_by_name['commandOptions'].message_type = _GETSERVERINFORESPONSE_COMMANDOPTIONINFO
_CLIENTCONNECTIONINFO_HTTPREQUESTINFO.containing_type = _CLIENTCONNECTIONINFO
_CLIENTCONNECTIONINFO.fields_by_name['httpRequest'].message_type = _CLIENTCONNECTIONINFO_HTTPREQUESTINFO
DESCRIPTOR.message_types_by_name['GetThreadRequest'] = _GETTHREADREQUEST
DESCRIPTOR.message_types_by_name['ListThreadsResponse'] = _LISTTHREADSRESPONSE
DESCRIPTOR.message_types_by_name['ListRoutesResponse'] = _LISTROUTESRESPONSE
DESCRIPTOR.message_types_by_name['ListTopicsResponse'] = _LISTTOPICSRESPONSE
DESCRIPTOR.message_types_by_name['HttpTraffic'] = _HTTPTRAFFIC
DESCRIPTOR.message_types_by_name['RouteInfo'] = _ROUTEINFO
DESCRIPTOR.message_types_by_name['TopicInfo'] = _TOPICINFO
DESCRIPTOR.message_types_by_name['TraceElementInfo'] = _TRACEELEMENTINFO
DESCRIPTOR.message_types_by_name['ListThreadsRequest'] = _LISTTHREADSREQUEST
DESCRIPTOR.message_types_by_name['ThreadGroupInfo'] = _THREADGROUPINFO
DESCRIPTOR.message_types_by_name['ThreadInfo'] = _THREADINFO
DESCRIPTOR.message_types_by_name['GetServerInfoResponse'] = _GETSERVERINFORESPONSE
DESCRIPTOR.message_types_by_name['ClientConnectionInfo'] = _CLIENTCONNECTIONINFO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetThreadRequest = _reflection.GeneratedProtocolMessageType('GetThreadRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETTHREADREQUEST,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.GetThreadRequest)
  ))
_sym_db.RegisterMessage(GetThreadRequest)

ListThreadsResponse = _reflection.GeneratedProtocolMessageType('ListThreadsResponse', (_message.Message,), dict(
  DESCRIPTOR = _LISTTHREADSRESPONSE,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ListThreadsResponse)
  ))
_sym_db.RegisterMessage(ListThreadsResponse)

ListRoutesResponse = _reflection.GeneratedProtocolMessageType('ListRoutesResponse', (_message.Message,), dict(
  DESCRIPTOR = _LISTROUTESRESPONSE,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ListRoutesResponse)
  ))
_sym_db.RegisterMessage(ListRoutesResponse)

ListTopicsResponse = _reflection.GeneratedProtocolMessageType('ListTopicsResponse', (_message.Message,), dict(
  DESCRIPTOR = _LISTTOPICSRESPONSE,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ListTopicsResponse)
  ))
_sym_db.RegisterMessage(ListTopicsResponse)

HttpTraffic = _reflection.GeneratedProtocolMessageType('HttpTraffic', (_message.Message,), dict(
  DESCRIPTOR = _HTTPTRAFFIC,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.HttpTraffic)
  ))
_sym_db.RegisterMessage(HttpTraffic)

RouteInfo = _reflection.GeneratedProtocolMessageType('RouteInfo', (_message.Message,), dict(
  DESCRIPTOR = _ROUTEINFO,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.RouteInfo)
  ))
_sym_db.RegisterMessage(RouteInfo)

TopicInfo = _reflection.GeneratedProtocolMessageType('TopicInfo', (_message.Message,), dict(
  DESCRIPTOR = _TOPICINFO,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.TopicInfo)
  ))
_sym_db.RegisterMessage(TopicInfo)

TraceElementInfo = _reflection.GeneratedProtocolMessageType('TraceElementInfo', (_message.Message,), dict(
  DESCRIPTOR = _TRACEELEMENTINFO,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.TraceElementInfo)
  ))
_sym_db.RegisterMessage(TraceElementInfo)

ListThreadsRequest = _reflection.GeneratedProtocolMessageType('ListThreadsRequest', (_message.Message,), dict(
  DESCRIPTOR = _LISTTHREADSREQUEST,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ListThreadsRequest)
  ))
_sym_db.RegisterMessage(ListThreadsRequest)

ThreadGroupInfo = _reflection.GeneratedProtocolMessageType('ThreadGroupInfo', (_message.Message,), dict(
  DESCRIPTOR = _THREADGROUPINFO,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ThreadGroupInfo)
  ))
_sym_db.RegisterMessage(ThreadGroupInfo)

ThreadInfo = _reflection.GeneratedProtocolMessageType('ThreadInfo', (_message.Message,), dict(
  DESCRIPTOR = _THREADINFO,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ThreadInfo)
  ))
_sym_db.RegisterMessage(ThreadInfo)

GetServerInfoResponse = _reflection.GeneratedProtocolMessageType('GetServerInfoResponse', (_message.Message,), dict(

  PluginInfo = _reflection.GeneratedProtocolMessageType('PluginInfo', (_message.Message,), dict(
    DESCRIPTOR = _GETSERVERINFORESPONSE_PLUGININFO,
    __module__ = 'yamcs.protobuf.web.server_service_pb2'
    # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.GetServerInfoResponse.PluginInfo)
    ))
  ,

  CommandOptionInfo = _reflection.GeneratedProtocolMessageType('CommandOptionInfo', (_message.Message,), dict(
    DESCRIPTOR = _GETSERVERINFORESPONSE_COMMANDOPTIONINFO,
    __module__ = 'yamcs.protobuf.web.server_service_pb2'
    # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.GetServerInfoResponse.CommandOptionInfo)
    ))
  ,
  DESCRIPTOR = _GETSERVERINFORESPONSE,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.GetServerInfoResponse)
  ))
_sym_db.RegisterMessage(GetServerInfoResponse)
_sym_db.RegisterMessage(GetServerInfoResponse.PluginInfo)
_sym_db.RegisterMessage(GetServerInfoResponse.CommandOptionInfo)

ClientConnectionInfo = _reflection.GeneratedProtocolMessageType('ClientConnectionInfo', (_message.Message,), dict(

  HttpRequestInfo = _reflection.GeneratedProtocolMessageType('HttpRequestInfo', (_message.Message,), dict(
    DESCRIPTOR = _CLIENTCONNECTIONINFO_HTTPREQUESTINFO,
    __module__ = 'yamcs.protobuf.web.server_service_pb2'
    # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ClientConnectionInfo.HttpRequestInfo)
    ))
  ,
  DESCRIPTOR = _CLIENTCONNECTIONINFO,
  __module__ = 'yamcs.protobuf.web.server_service_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.ClientConnectionInfo)
  ))
_sym_db.RegisterMessage(ClientConnectionInfo)
_sym_db.RegisterMessage(ClientConnectionInfo.HttpRequestInfo)


DESCRIPTOR._options = None

_SERVERAPI = _descriptor.ServiceDescriptor(
  name='ServerApi',
  full_name='yamcs.protobuf.web.ServerApi',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=2187,
  serialized_end=2897,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetServerInfo',
    full_name='yamcs.protobuf.web.ServerApi.GetServerInfo',
    index=0,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_GETSERVERINFORESPONSE,
    serialized_options=_b('\212\222\003\006\n\004/api'),
  ),
  _descriptor.MethodDescriptor(
    name='ListRoutes',
    full_name='yamcs.protobuf.web.ServerApi.ListRoutes',
    index=1,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_LISTROUTESRESPONSE,
    serialized_options=_b('\212\222\003\r\n\013/api/routes'),
  ),
  _descriptor.MethodDescriptor(
    name='ListTopics',
    full_name='yamcs.protobuf.web.ServerApi.ListTopics',
    index=2,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_LISTTOPICSRESPONSE,
    serialized_options=_b('\212\222\003\r\n\013/api/topics'),
  ),
  _descriptor.MethodDescriptor(
    name='ListThreads',
    full_name='yamcs.protobuf.web.ServerApi.ListThreads',
    index=3,
    containing_service=None,
    input_type=_LISTTHREADSREQUEST,
    output_type=_LISTTHREADSRESPONSE,
    serialized_options=_b('\212\222\003\016\n\014/api/threads'),
  ),
  _descriptor.MethodDescriptor(
    name='GetThread',
    full_name='yamcs.protobuf.web.ServerApi.GetThread',
    index=4,
    containing_service=None,
    input_type=_GETTHREADREQUEST,
    output_type=_THREADINFO,
    serialized_options=_b('\212\222\003\023\n\021/api/threads/{id}'),
  ),
  _descriptor.MethodDescriptor(
    name='DumpThreads',
    full_name='yamcs.protobuf.web.ServerApi.DumpThreads',
    index=5,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=yamcs_dot_api_dot_httpbody__pb2._HTTPBODY,
    serialized_options=_b('\212\222\003\023\n\021/api/threads:dump'),
  ),
  _descriptor.MethodDescriptor(
    name='GetHttpTraffic',
    full_name='yamcs.protobuf.web.ServerApi.GetHttpTraffic',
    index=6,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=_HTTPTRAFFIC,
    serialized_options=_b('\212\222\003\023\n\021/api/http-traffic'),
  ),
])
_sym_db.RegisterServiceDescriptor(_SERVERAPI)

DESCRIPTOR.services_by_name['ServerApi'] = _SERVERAPI

# @@protoc_insertion_point(module_scope)



import threading

from yamcs.core.helpers import parse_server_time
from yamcs.protobuf.filetransfer import filetransfer_pb2


class Service:
    def __init__(self, proto, service_client):
        self._proto = proto
        self._service_client = service_client

    @property
    def name(self):
        """Name of this service."""
        return self._proto.name

    def upload(
        self,
        bucket_name,
        object_name,
        remote_path,
        source_entity=None,
        destination_entity=None,
        overwrite=True,
        parents=True,
        reliable=False,
    ):
        """
        Uploads a file located in a bucket to a remote destination path.

        :param str bucket_name: Name of the bucket containing the source object.
        :param str object_name: Name of the source object.
        :param str remote_path: Remote destination.
        :param str source_entity: Use a specific source entity.
                                  (useful in case of multiples)
        :param str destination_entity: Use a specific destination entity.
                                       (useful in case of multiples)
        :param bool overwrite: Replace a destination if it already exists.
        :param bool parents: Create the remote path if it does not yet exist.
        :param bool reliable: Whether to use a Class 2 CFDP transfer.
        :rtype: .Transfer
        """
        return self._service_client.upload(
            bucket_name=bucket_name,
            object_name=object_name,
            remote_path=remote_path,
            source_entity=source_entity,
            destination_entity=destination_entity,
            overwrite=overwrite,
            parents=parents,
            reliable=reliable,
        )

    def download(
        self,
        bucket_name,
        object_name,
        remote_path,
        source_entity=None,
        destination_entity=None,
        overwrite=True,
        parents=True,
        reliable=False,
    ):
        """
        Downloads a file from the destination to a bucket.

        :param str bucket_name: Name of the bucket to receive the file.
        :param str object_name: Name of the file received in the bucket.
        :param str remote_path: Name of the file to be downloaded from the destination.
        :param str source_entity: Use a specific source entity.
                                  (useful in case of multiples)
        :param str destination_entity: Use a specific destination entity.
                                       (useful in case of multiples)
        :param bool overwrite: Replace file if it already exists.
        :param bool parents: Create the remote path if it does not yet exist.
        :param bool reliable: Enable reliable transfers.
        :rtype: .Transfer
        """
        return self._service_client.download(
            bucket_name=bucket_name,
            object_name=object_name,
            remote_path=remote_path,
            source_entity=source_entity,
            destination_entity=destination_entity,
            overwrite=overwrite,
            parents=parents,
            reliable=reliable,
        )

    def pause_transfer(self, id):
        """
        Pauses a transfer
        """
        self._service_client.pause_transfer(id)

    def resume_transfer(self, id):
        """
        Resume a transfer
        """
        self._service_client.resume_transfer(id)

    def cancel_transfer(self, id):
        """
        Cancel a transfer
        """
        self._service_client.cancel_transfer(id)

    def create_transfer_subscription(self, on_data=None, timeout=60):
        """
        Create a new transfer subscription.

        :param on_data: (Optional) Function that gets called with
                        :class:`.TransferInfo` updates.
        :param timeout: The amount of seconds to wait for the request to
                        complete.
        :type timeout: float
        :return: Future that can be used to manage the background websocket
                 subscription
        :rtype: .TransferSubscription
        """
        return self._service_client.create_transfer_subscription(
            on_data=on_data, timeout=timeout
        )

    def __str__(self):
        return self.name


class Transfer:
    """
    Represents a file transfer.
    """

    def __init__(self, proto, service_client):
        self._proto = proto
        self._service_client = service_client

    @property
    def id(self):
        """Yamcs-local transfer identifier."""
        return self._proto.id

    @property
    def bucket(self):
        return self._proto.bucket

    @property
    def object_name(self):
        return self._proto.objectName

    @property
    def remote_path(self):
        return self._proto.remotePath

    @property
    def time(self):
        """Time when the transfer was started."""
        if self._proto.HasField("startTime"):
            return parse_server_time(self._proto.startTime)
        return None

    @property
    def reliable(self):
        """True if this is a Class 2 CFDP transfer."""
        return self._proto.reliable

    @property
    def state(self):
        """Current transfer state."""
        if self._proto.HasField("state"):
            return filetransfer_pb2.TransferState.Name(self._proto.state)
        return None

    @property
    def size(self):
        """Total bytes to transfer."""
        return self._proto.totalSize

    @property
    def transferred_size(self):
        """Total bytes already transferred."""
        return self._proto.sizeTransferred

    def is_complete(self):
        """
        Returns whether this transfer is complete. A transfer
        can be completed, yet still failed.
        """
        return self.state == "FAILED" or self.state == "COMPLETED"

    def is_success(self):
        """
        Returns true if this transfer was completed successfully.
        """
        return self.state == "COMPLETED"

    @property
    def error(self):
        """Error message in case the transfer failed."""
        if self.state == "FAILED" and self._proto.HasField("failureReason"):
            return self._proto.failureReason
        return None

    def pause(self):
        """
        Pause this transfer
        """
        self._service_client.pause_transfer(self.id)

    def resume(self):
        """
        Resume this transfer
        """
        self._service_client.resume_transfer(self.id)

    def cancel(self):
        """
        Cancel this transfer
        """
        self._service_client.cancel_transfer(self.id)

    def await_complete(self, timeout=None):
        """
        Wait for the transfer to be completed.

        :param float timeout: The amount of seconds to wait.
        """
        completed = threading.Event()

        def callback(updated_transfer):
            if updated_transfer.id == self.id:
                self._proto = updated_transfer._proto
                if self.is_complete():
                    completed.set()

        self._service_client.create_transfer_subscription(on_data=callback)

        if not completed.wait(timeout=timeout):
            # Remark that a timeout does *not* mean that the underlying
            # work is canceled.
            raise TimeoutError("Timed out.")



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: yamcs/protobuf/web/auth.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from yamcs.protobuf.iam import iam_pb2 as yamcs_dot_protobuf_dot_iam_dot_iam__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='yamcs/protobuf/web/auth.proto',
  package='yamcs.protobuf.web',
  syntax='proto2',
  serialized_options=_b('\n\022org.yamcs.protobufB\tAuthProtoP\001'),
  serialized_pb=_b('\n\x1dyamcs/protobuf/web/auth.proto\x12\x12yamcs.protobuf.web\x1a\x1cyamcs/protobuf/iam/iam.proto\"p\n\x08\x41uthInfo\x12\x1d\n\x15requireAuthentication\x18\x01 \x01(\x08\x12\x0e\n\x06spnego\x18\x02 \x01(\x08\x12\x35\n\x06openid\x18\x03 \x01(\x0b\x32%.yamcs.protobuf.web.OpenIDConnectInfo\"S\n\x11OpenIDConnectInfo\x12\x10\n\x08\x63lientId\x18\x01 \x01(\t\x12\x1d\n\x15\x61uthorizationEndpoint\x18\x02 \x01(\t\x12\r\n\x05scope\x18\x03 \x01(\t\"\x90\x01\n\rTokenResponse\x12\x14\n\x0c\x61\x63\x63\x65ss_token\x18\x01 \x01(\t\x12\x12\n\ntoken_type\x18\x02 \x01(\t\x12\x12\n\nexpires_in\x18\x03 \x01(\x05\x12\x15\n\rrefresh_token\x18\x04 \x01(\t\x12*\n\x04user\x18\x05 \x01(\x0b\x32\x1c.yamcs.protobuf.iam.UserInfoB!\n\x12org.yamcs.protobufB\tAuthProtoP\x01')
  ,
  dependencies=[yamcs_dot_protobuf_dot_iam_dot_iam__pb2.DESCRIPTOR,])




_AUTHINFO = _descriptor.Descriptor(
  name='AuthInfo',
  full_name='yamcs.protobuf.web.AuthInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='requireAuthentication', full_name='yamcs.protobuf.web.AuthInfo.requireAuthentication', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spnego', full_name='yamcs.protobuf.web.AuthInfo.spnego', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='openid', full_name='yamcs.protobuf.web.AuthInfo.openid', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=83,
  serialized_end=195,
)


_OPENIDCONNECTINFO = _descriptor.Descriptor(
  name='OpenIDConnectInfo',
  full_name='yamcs.protobuf.web.OpenIDConnectInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='clientId', full_name='yamcs.protobuf.web.OpenIDConnectInfo.clientId', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='authorizationEndpoint', full_name='yamcs.protobuf.web.OpenIDConnectInfo.authorizationEndpoint', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='scope', full_name='yamcs.protobuf.web.OpenIDConnectInfo.scope', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=197,
  serialized_end=280,
)


_TOKENRESPONSE = _descriptor.Descriptor(
  name='TokenResponse',
  full_name='yamcs.protobuf.web.TokenResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='access_token', full_name='yamcs.protobuf.web.TokenResponse.access_token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='token_type', full_name='yamcs.protobuf.web.TokenResponse.token_type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='expires_in', full_name='yamcs.protobuf.web.TokenResponse.expires_in', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='refresh_token', full_name='yamcs.protobuf.web.TokenResponse.refresh_token', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='user', full_name='yamcs.protobuf.web.TokenResponse.user', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=283,
  serialized_end=427,
)

_AUTHINFO.fields_by_name['openid'].message_type = _OPENIDCONNECTINFO
_TOKENRESPONSE.fields_by_name['user'].message_type = yamcs_dot_protobuf_dot_iam_dot_iam__pb2._USERINFO
DESCRIPTOR.message_types_by_name['AuthInfo'] = _AUTHINFO
DESCRIPTOR.message_types_by_name['OpenIDConnectInfo'] = _OPENIDCONNECTINFO
DESCRIPTOR.message_types_by_name['TokenResponse'] = _TOKENRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AuthInfo = _reflection.GeneratedProtocolMessageType('AuthInfo', (_message.Message,), dict(
  DESCRIPTOR = _AUTHINFO,
  __module__ = 'yamcs.protobuf.web.auth_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.AuthInfo)
  ))
_sym_db.RegisterMessage(AuthInfo)

OpenIDConnectInfo = _reflection.GeneratedProtocolMessageType('OpenIDConnectInfo', (_message.Message,), dict(
  DESCRIPTOR = _OPENIDCONNECTINFO,
  __module__ = 'yamcs.protobuf.web.auth_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.OpenIDConnectInfo)
  ))
_sym_db.RegisterMessage(OpenIDConnectInfo)

TokenResponse = _reflection.GeneratedProtocolMessageType('TokenResponse', (_message.Message,), dict(
  DESCRIPTOR = _TOKENRESPONSE,
  __module__ = 'yamcs.protobuf.web.auth_pb2'
  # @@protoc_insertion_point(class_scope:yamcs.protobuf.web.TokenResponse)
  ))
_sym_db.RegisterMessage(TokenResponse)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)
